#include <Box.h>
#include "common.h"
#include "ctrl.h"


VolMenuItem::VolMenuItem(BVolume *v, BMessage *m):
    BMenuItem(kEmptyStr, m), _volume(v)
{
    _volume->GetName(strbuf);
    SetLabel(strbuf);

    SetLabel(strbuf);

    _icon = new BBitmap(BRect(0.0, 0.0, 15.0, 15.0), B_CMAP8);
    _volume->GetIcon(_icon, B_MINI_ICON);
}


VolMenuItem::~VolMenuItem()
{
    delete _icon;
    delete _volume;
}


void
VolMenuItem::GetContentSize(float *width, float *height)
{
    *width = _icon->Bounds().Width() +
	be_plain_font->StringWidth(Label());

    struct font_height fh;
    be_plain_font->GetHeight(&fh);
    float fontHeight = fh.ascent + fh.descent + fh.leading;
    *height = max(fontHeight, _icon->Bounds().Height());
}


void
VolMenuItem::DrawContent()
{
    Menu()->SetDrawingMode(B_OP_OVER);
    Menu()->MovePenBy(0.0, -1.0);
    Menu()->DrawBitmap(_icon);
    Menu()->SetDrawingMode(B_OP_COPY);
    Menu()->MovePenBy(_icon->Bounds().Width() + kSmallHMargin, 0.0);
    BMenuItem::DrawContent();
}


VolCtrl::VolCtrl(BRect r):
    BMenuField(r, NULL, kVolMenuLabel, new BPopUpMenu(kVolMenuDefault),
	false, B_FOLLOW_LEFT)
{
    SetViewColor(kWindowColor);
    SetLowColor(kWindowColor);

    SetDivider(kSmallHMargin + StringWidth(kVolMenuLabel));

    // Populate the menu with the persistent volumes.
    _roster = new BVolumeRoster();
    BVolume vol, bootvol;
    _roster->GetBootVolume(&bootvol);

    while (_roster->GetNextVolume(&vol) == B_NO_ERROR) {
	if (vol.IsPersistent()) {
	    BVolume *v = new BVolume(vol);
	    BMessage *m = new BMessage(kMenuSelectVol);
	    m->AddPointer(kNameVolPtr, v);
	    VolMenuItem *item = new VolMenuItem(v, m);
	    Menu()->AddItem(item);
	}
    }
}


VolCtrl::~VolCtrl()
{
    _roster->StopWatching();
    delete _roster;
}


void
VolCtrl::_addVolume(dev_t device)
{
    // Make sure the volume is not already in the menu.
    for (int i = 0; i < Menu()->CountItems(); i++) {
	BVolume *menuVol = ((VolMenuItem *)Menu()->ItemAt(i))->volume();
	if (menuVol->Device() == device) {
	    return;
	}
    }

    // Add the newly mounted volume to the menu.
    BVolume *v = new BVolume(device);
    BMessage *m = new BMessage(kMenuSelectVol);
    m->AddPointer(kNameVolPtr, v);
    VolMenuItem *item = new VolMenuItem(v, m);
    Menu()->AddItem(item);
}


void
VolCtrl::_removeVolume(dev_t device)
{
    for (int i = 0; i < Menu()->CountItems(); i++) {
	BMenuItem *item = Menu()->ItemAt(i);
	if (((VolMenuItem *)item)->volume()->Device() == device) {
	    // If the volume being removed is currently selected, prompt for a
	    // different volume.
	    if (item->IsMarked()) {
		// BUG: how to update the display now that there is no marked
		// item?  This doesn't work:
		//item->SetLabel(kVolMenuDefault);
		BMessage m(kMenuSelectVol);
		m.AddPointer(kNameVolPtr, NULL);
		Window()->PostMessage(&m);
	    }

	    Menu()->RemoveItem(item);
	    return;
	}
    }
}


BVolume *
VolCtrl::findDevice(dev_t device, bool invoke)
{
    BVolume *vol = NULL;

    // Iterate thru items looking for a BVolume representing this device.
    for (int i = 0; true; i++) {
	VolMenuItem *item = (VolMenuItem *) Menu()->ItemAt(i);
	if (item == NULL) {
	    break;
	}
	if (item->volume()->Device() == device) {
	    vol = item->volume();
	    if (invoke) {
		item->invoke();
	    }
	    break;
	}
    }

    return vol;
}


void
VolCtrl::AttachedToWindow()
{
    // Begin watching mount and unmount events.
    _roster->StartWatching(BMessenger(this));
}


void
VolCtrl::MessageReceived(BMessage *m)
{
    switch (m->what) {
	case B_NODE_MONITOR:
	    switch (m->FindInt32("opcode")) {
		case B_DEVICE_MOUNTED:
		    _addVolume(m->FindInt32("new device"));
		    break;

		case B_DEVICE_UNMOUNTED:
		    _removeVolume(m->FindInt32("device"));
		    break;
	    }
	    break;

        default:
            BMenuField::MessageReceived(m);
            break;
    }
}


CtrlView::CtrlView(BRect r):
    BView(r, NULL, B_FOLLOW_LEFT_RIGHT, 0)
{
    SetViewColor(kWindowColor);

    // BUG: adding these right-to-left messes up keyboard navigation.

    r.top += kSmallVMargin;
    r.right -= kSmallHMargin;
    float buttonWidth = kButtonMargin + StringWidth(kHelpBtnLabel);
    //r.left = r.right - max(kMinButtonWidth, buttonWidth);
    r.left = r.right - buttonWidth;
    BButton *b = new BButton(r, NULL, kHelpBtnLabel, new BMessage(kBtnHelp),
	B_FOLLOW_RIGHT);
    if (!kFoundHelpFile) {
	b->SetEnabled(false);
    }
    AddChild(b);

    r.right = r.left - kSmallHMargin;
    buttonWidth = kButtonMargin + StringWidth(kStrRescan);
    r.left = r.right - max(kMinButtonWidth, buttonWidth);
    _rescanButton = new BButton(r, NULL, kStrRescan, new BMessage(kBtnRescan),
	B_FOLLOW_RIGHT);
    AddChild(_rescanButton);

    volCtrl = new VolCtrl(
	BRect(kSmallHMargin, kSmallVMargin, r.left - kSmallHMargin, kSmallVMargin));
    AddChild(volCtrl);

    float width, height;
    _rescanButton->GetPreferredSize(&width, &height);
    ResizeTo(Bounds().Width(), height + 6.0);

    // Horizontal divider
    BRect r = Bounds();
    r.top = r.bottom - 1.0;
    r.left -= 5.0; r.right += 5.0;
    AddChild(new BBox(r, NULL, B_FOLLOW_LEFT_RIGHT, B_WILL_DRAW));
}


CtrlView::~CtrlView()
{
}

