#include <File.h>
#include <AppFileInfo.h>
#include <Path.h>
#include <math.h>
#include <fs_info.h>
#include "pie.h"
#include "ctrl.h"
#include "status.h"
#include "info.h"
#include "common.h" // #included last to get my definition of min()


const int32 kIdxGetInfo = 0;
const int32 kIdxOpen = 1;
const int32 kIdxOpenWith = 2;
const int32 kIdxRescan = 3;

// Some time it would be nice to make a common base class for AppMenuItem and
// VolMenuItem (menu items that include an icon).

AppMenuItem::AppMenuItem(const char *appSig, int category):
    BMenuItem(kEmptyStr, NULL), _category(category), _icon(NULL), _valid(false)
{
    if (be_roster->FindApp(appSig, &_appRef) == B_NO_ERROR) {
	_icon = new BBitmap(BRect(0.0, 0.0, 15.0, 15.0), B_CMAP8);
	if (BNodeInfo::GetTrackerIcon(&_appRef, _icon, B_MINI_ICON) == B_OK) {
	    BEntry appEntry(&_appRef);
	    if (appEntry.InitCheck() == B_OK) {
		appEntry.GetName(strbuf);
		SetLabel(strbuf);
		_valid = true;
	    }
	}
    }
}


AppMenuItem::~AppMenuItem()
{
    if (_icon != NULL) {
	delete _icon;
    }
}


void
AppMenuItem::GetContentSize(float *width, float *height)
{
    *width = _icon->Bounds().Width() +
	be_plain_font->StringWidth(Label());

    struct font_height fh;
    be_plain_font->GetHeight(&fh);
    float fontHeight = fh.ascent + fh.descent + fh.leading;
    *height = max(fontHeight, _icon->Bounds().Height());
}


void
AppMenuItem::DrawContent()
{
    Menu()->SetDrawingMode(B_OP_OVER);
    Menu()->MovePenBy(0.0, -1.0);
    Menu()->DrawBitmap(_icon);
    Menu()->MovePenBy(_icon->Bounds().Width() + kSmallHMargin, 0.0);
    BMenuItem::DrawContent();
}


PieView::PieView(BRect r):
    BView(r, NULL, B_FOLLOW_ALL,
	B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE|B_SUBPIXEL_PRECISE),
    _scanners(), _currentVolume(NULL), _mouseOverInfo(), _clicked(false),
    _dragging(false)
{
    SetViewColor(kWindowColor);
    SetLowColor(kWindowColor);
    SetDrawingMode(B_OP_COPY);

    _mouseOverMenu = new BPopUpMenu(kEmptyStr, false, false);
    _mouseOverMenu->AddItem(new BMenuItem(kMenuGetInfo, NULL), kIdxGetInfo);
    _mouseOverMenu->AddItem(new BMenuItem(kMenuOpen, NULL), kIdxOpen);

    _fileUnavailMenu = new BPopUpMenu(kEmptyStr, false, false);
    BMenuItem *item = new BMenuItem(kStrUnavail, NULL);
    item->SetEnabled(false);
    _fileUnavailMenu->AddItem(item);

    BFont f;
    GetFont(&f);
    f.SetSize(16.0);
    f.SetFace(B_BOLD_FACE);
    SetFont(&f);

    struct font_height fh;
    f.GetHeight(&fh);
    _fontHeight = fh.ascent + fh.descent + fh.leading;
}


PieView::~PieView()
{
    delete _mouseOverMenu;
    delete _fileUnavailMenu;

    while (_scanners.size() != 0) {
	ScannerMap::iterator i = _scanners.begin();
	(*i).second->requestQuit();
	_scanners.erase(i);
    }
}


void
PieView::_showVolume(BVolume *v)
{
    if (v != NULL) {
	if (_scanners.find(v) == _scanners.end()) {
	    _scanners[v] = new Scanner(v, this);
	}

	Scanner *scanner = _scanners[v];
	if (scanner->snapshot() == NULL) {
	    scanner->refresh();
	}
    }

    _currentVolume = v;
    Invalidate();
}


void
PieView::_drawProgressBar(BRect updateRect)
{
    // Show the progress of the scanning operation.

    _mouseOverInfo.clear();

    // Draw the progress bar.
    BRect b = Bounds();
    float bx = (b.left + b.Width() - kProgBarWidth) / 2.0;
    float by = (b.top + b.Height() - kProgBarHeight) / 2.0;
    float ex = bx + kProgBarWidth;
    float ey = by + kProgBarHeight;
    SetPenSize(0.0);
    SetHighColor(0xf0, 0xf0, 0xf0);
    StrokeLine(BPoint(bx, ey), BPoint(ex, ey));
    StrokeLine(BPoint(ex, by), BPoint(ex, ey));
    SetHighColor(0x40, 0x40, 0x40);
    StrokeLine(BPoint(bx, by), BPoint(ex, by));
    StrokeLine(BPoint(bx, by), BPoint(bx, ey));

    bx += 1.0; by += 1.0;
    ex -= 1.0; ey -= 1.0;
    float mx = bx + (kProgBarWidth - 2.0) *
	_scanners[_currentVolume]->progress() / 100.0;
    SetHighColor(kWindowColor);
    FillRect(BRect(mx, by, ex, ey));

    SetHighColor(0x00, 0x20, 0x90);
    StrokeLine(BPoint(bx, ey), BPoint(mx, ey));
    StrokeLine(BPoint(mx, by), BPoint(mx, ey));
    SetHighColor(0x00, 0xc0, 0xff);
    StrokeLine(BPoint(bx, by), BPoint(mx, by));
    StrokeLine(BPoint(bx, by), BPoint(bx, ey));

    bx += 1.0; by += 1.0;
    mx -= 1.0; ey -= 1.0;
    SetHighColor(0x00, 0x80, 0xf0);
    FillRect(BRect(bx, by, mx, ey));

    // Tell what we are doing.
    const char *task = _scanners[_currentVolume]->task();
    float strWidth = StringWidth(task);
    bx = (b.left + b.Width() - strWidth) / 2.0;
    by -= _fontHeight + 2.0 * kSmallVMargin;
    SetHighColor(0x00, 0x00, 0x00); // BUG: no hard-code
    DrawString(task, BPoint(bx, by));
}


void
PieView::_drawPieChart(BRect updateRect)
{
    BRect pieRect = Bounds();
    pieRect.InsetBy(kPieOuterMargin, kPieOuterMargin);
    if (updateRect.Intersects(pieRect)) {
	SetHighColor(kPieBGColor);
	FillEllipse(pieRect);

	int colorIdx = 0;
	FileInfo *currentDir = _scanners[_currentVolume]->currentDir();
	FileInfo *f = currentDir;
	while (f != NULL) {
	    f = f->parent;
	    colorIdx++;
	}
	_drawDirectory(currentDir, 0.0, 0.0, colorIdx % kBasePieColorCount, 0);

	BPoint p;
	uint32 ignore;
	GetMouse(&p, &ignore, false);
	_win->statusView->show(_fileAt(p));
    }
}


float
PieView::_drawDirectory(FileInfo *f, float parentSpan, float beginAngle, int
			colorIdx, int level)
{
    BRect b = Bounds();
    if (b.Width() < 2.0 * (kPieCenterSize + level * kPieRingSize +
	    kPieOuterMargin + kPieInnerMargin)) {
	return 0.0;
    }

    VolSnapshot *snapshot = _scanners[_currentVolume]->snapshot();

    float cx = b.left + b.Width() / 2.0;
    float cy = b.top + b.Height() / 2.0;

    float mySpan;

    if (level == 0) {
	// Make room for mouse over info.
	_mouseOverInfo.clear();
	_mouseOverInfo[0] = SegmentList();

	// Draw the center circle.
	const char *displayName;
	if (f == NULL) {
	    // NULL represents the entire volume.  Show used and free space in
	    // the center circle, with the used segment representing the
	    // volume's root directory.
	    off_t volCapacity = snapshot->capacity;
	    mySpan = 360.0 *
		(volCapacity - snapshot->freeBytes) / volCapacity;

	    SetHighColor(kEmptySpcColor);
	    FillEllipse(BPoint(cx, cy), kPieCenterSize, kPieCenterSize);

	    SetHighColor(kBasePieColor[0]);
	    FillArc(BPoint(cx, cy), kPieCenterSize, kPieCenterSize, 0.0, mySpan);

	    // Show total volume capacity.
	    const char *c = scaleByteCount(volCapacity);
	    SetHighColor(kPieBGColor);
	    SetDrawingMode(B_OP_OVER);
	    DrawString(c,
		BPoint(cx - StringWidth(c)/2.0, cy + _fontHeight + kSmallVMargin));
	    SetDrawingMode(B_OP_COPY);
	    displayName = snapshot->name.c_str();

	    // Record in-use space and free space for use during MouseMoved().
	    f = snapshot->rootDir;
	    _mouseOverInfo[0].push_back(Segment(0.0, mySpan, f));
	    if (mySpan < 360.0 - kMinSegmentSpan) {
		_mouseOverInfo[0].push_back(Segment(mySpan, 360.0, snapshot->freeSpace));
	    }
	}
	else {
	    // Show a normal directory.
	    SetHighColor(kBasePieColor[colorIdx]);
	    FillEllipse(BPoint(cx, cy), kPieCenterSize, kPieCenterSize);
	    displayName = f->ref.name;
	    mySpan = 360.0;

	    // Record the segment for use during MouseMoved().
	    _mouseOverInfo[0].push_back(Segment(0.0, mySpan, f));
	}

	SetPenSize(0.0);
	SetHighColor(kOutlineColor);
	StrokeEllipse(BPoint(cx, cy), kPieCenterSize, kPieCenterSize);

	// Show the name of the volume or directory.
	char *dncopy = strdup(displayName);
	float sw = StringWidth(dncopy);

	// If the name is too long to fit in the center circle, truncate it
	// and an an ellipsis.
	if (sw > 2.0 * (kPieCenterSize - kSmallHMargin)) {
	    float ew = StringWidth(B_UTF8_ELLIPSIS);
	    char *cptr = dncopy + strlen(dncopy) - strlen(B_UTF8_ELLIPSIS);
	    while (sw + ew > 2.0 * (kPieCenterSize - kSmallHMargin)) {
		*cptr = '\0';
		cptr--;
		sw = StringWidth(dncopy);
	    }
	    strcpy(cptr, B_UTF8_ELLIPSIS);
	    sw = StringWidth(dncopy);
	}

	SetHighColor(kPieBGColor);
	SetDrawingMode(B_OP_OVER);
	DrawString(dncopy, BPoint(cx - sw/2.0, cy));
	SetDrawingMode(B_OP_COPY);
	free(dncopy);
	beginAngle = 0.0;
    }
    else {
	// Draw an exterior segment.
	float parentSize;
	if (f->parent == NULL) {
	    parentSize = (float)snapshot->capacity;
	}
	else {
	    parentSize = (float)f->parent->size;
	}

	mySpan = parentSpan * (float)f->size / parentSize;
	if (mySpan >= kMinSegmentSpan) {
	    rgb_color color = kBasePieColor[colorIdx];
	    color.red += kLightenFactor * level;
	    color.green += kLightenFactor * level;
	    color.blue += kLightenFactor * level;
	    SetHighColor(color);
	    SetPenSize(kPieRingSize);
	    float radius = kPieCenterSize + level * kPieRingSize - kPieRingSize / 2.0;
	    StrokeArc(BPoint(cx, cy), radius, radius, beginAngle, mySpan);

	    SetHighColor(kOutlineColor);
	    SetPenSize(0.0);
	    float segBeginRadius = kPieCenterSize + (level - 1) * kPieRingSize + 0.5;
	    float segLength = kPieRingSize - 0.5;
	    float rad = deg2rad(beginAngle);
	    float bx = cx + segBeginRadius * cos(rad);
	    float by = cy - segBeginRadius * sin(rad);
	    float ex = bx + segLength * cos(rad);
	    float ey = by - segLength * sin(rad);
	    StrokeLine(BPoint(bx, by), BPoint(ex, ey));

	    rad = deg2rad(beginAngle + mySpan);
	    bx = cx + segBeginRadius * cos(rad);
	    by = cy - segBeginRadius * sin(rad);
	    ex = bx + segLength * cos(rad);
	    ey = by - segLength * sin(rad);
	    StrokeLine(BPoint(bx, by), BPoint(ex, ey));

	    SetPenSize(0.0);
	    SetHighColor(kOutlineColor);
	    radius += kPieRingSize / 2.0;
	    StrokeArc(BPoint(cx, cy), radius, radius, beginAngle, mySpan);

	    // Record the segment for use during MouseMoved().
	    if (_mouseOverInfo.find(level) == _mouseOverInfo.end()) {
		_mouseOverInfo[level] = SegmentList();
	    }
	    _mouseOverInfo[level].push_back(
		Segment(beginAngle, beginAngle + mySpan, f));
	}
    }

    // Draw children.
    vector<FileInfo *>::iterator i = f->children.begin();
    while (i != f->children.end()) {
	float childSpan =
	    _drawDirectory(*i, mySpan, beginAngle, colorIdx, level + 1);
	if (childSpan >= kMinSegmentSpan) {
	    beginAngle += childSpan;
	    colorIdx = (colorIdx + 1) % kBasePieColorCount;
	}
	i++;
    }

    return mySpan;
}


FileInfo *
PieView::_fileAt(BPoint p)
{
    BRect b = Bounds();
    float cx = b.left + b.Width() / 2.0;
    float cy = b.top + b.Height() / 2.0;
    float dx = p.x - cx;
    float dy = p.y - cy;
    float dist = sqrt(dx*dx + dy*dy);

    int level;
    if (dist < kPieCenterSize) {
	level = 0;
    }
    else {
	level = 1 + (int)((dist - kPieCenterSize) / kPieRingSize);
    }

    float angle = rad2deg(atan(dy/dx));
    angle = ((dx < 0.0) ? 180.0 : (dy < 0.0) ? 0.0 : 360.0) - angle;

    if (_mouseOverInfo.find(level) == _mouseOverInfo.end()) {
	// No files in this level (ring) of the pie.
	return NULL;
    }

    SegmentList s = _mouseOverInfo[level];
    SegmentList::iterator i = s.begin();
    while (i != s.end() &&
	    (angle < (*i).begin || (*i).end < angle)) {
	i++;
    }
    if (i == s.end()) {
	// Nothing at this angle.
	return NULL;
    }
    
    return (*i).info;
}


void
PieView::_addAppToList(vector<AppMenuItem*> &list, const char *appSig, int category)
{
    // Skip me.
    if (strcmp(appSig, kGuidoSig) == 0) {
	return;
    }

    AppMenuItem *item = new AppMenuItem(appSig, category);
    if (item->valid()) {
	vector<AppMenuItem*>::iterator i = list.begin();
	while (i != list.end()) {
	    if (*item->appRef() == *(*i)->appRef()) {
		// Skip duplicates.
		delete item;
		return;
	    }
	    i++;
	}
	list.push_back(item);
    }
    else {
	// Skip items that weren't constructed successfully.
	delete item;
    }
}


BMenu *
PieView::_buildOpenWithMenu(FileInfo *f)
{
    vector<AppMenuItem*> appList;

    // Get preferred app.
    BMimeType *type = f->type();
    if (type->GetPreferredApp(strbuf) == B_OK) {
	_addAppToList(appList, strbuf, 1);
    }

    // Get apps that handle this subtype and supertype.
    BMessage msg;
    if (type->GetSupportingApps(&msg) == B_OK) {
	int32 subs, supers, i;
	// BUG: report doc bug: sample code sez "be:subs" &c.
	msg.FindInt32("be:sub", &subs);
	msg.FindInt32("be:super", &supers);

	const char *appSig;
	for (i = 0; i < subs; i++) {
	    msg.FindString("applications", i, &appSig);
	    _addAppToList(appList, appSig, 2);
	}
	int hold = i;
	for (i = 0; i < supers; i++) {
	    msg.FindString("applications", i + hold, &appSig);
	    _addAppToList(appList, appSig, 3);
	}
    }

    // Get apps that handle any type.
    if (BMimeType::GetWildcardApps(&msg) == B_OK) {
	const char *appSig;
	for (int32 i = 0; true; i++) {
	    if (msg.FindString("applications", i, &appSig) == B_OK) {
		_addAppToList(appList, appSig, 4);
	    }
	    else {
		break;
	    }
	}
    }

    delete type;

    BMenu *openWith = new BMenu(kMenuOpenWith);

    if (appList.size() == 0) {
	BMenuItem *item = new BMenuItem(kMenuNoApps, NULL);
	item->SetEnabled(false);
	openWith->AddItem(item);
    }
    else {
	vector<AppMenuItem*>::iterator i = appList.begin();
	int category = (*i)->category();
	while (i != appList.end()) {
	    if (category != (*i)->category()) {
		openWith->AddSeparatorItem();
		category = (*i)->category();
	    }
	    openWith->AddItem(*i);
	    i++;
	}
    }

    return openWith;
}


void
PieView::_menu(FileInfo *f, BPoint p)
{
    BRect openRect(p.x - 5.0, p.y - 5.0, p.x + 5.0, p.y + 5.0);

    // Display the open-with menu only if the file is still available.
    BNode node(&f->ref);
    if (node.InitCheck() == B_OK) {
	// Add "Open With" submenu.
	BMenu *openWith = _buildOpenWithMenu(f);
	_mouseOverMenu->AddItem(openWith, kIdxOpenWith);

	// Add a "Rescan" option for folders.
	BMenuItem *rescan = NULL;
	if (f->children.size() > 0) {
	    rescan = new BMenuItem(kStrRescan, NULL);
	    _mouseOverMenu->AddItem(rescan, kIdxRescan);
	}

	BMenuItem *item = _mouseOverMenu->Go(p, false, true, openRect);
	if (item != NULL) {
	    switch (_mouseOverMenu->IndexOf(item)) {
		case kIdxGetInfo:
		    new InfoWin(p, f);
		    break;
		case kIdxOpen:
		    _launch(f);
		    break;
		case kIdxRescan:
		    _scanners[_currentVolume]->refresh(f);
		    Invalidate();
		    break;
		default: // must be "Open With" submenu
		    _launch(f, ((AppMenuItem *)item)->appRef());
		    break;
	    }
	}

	if (rescan != NULL) {
	    _mouseOverMenu->RemoveItem(rescan);
	    delete rescan;
	}

	_mouseOverMenu->RemoveItem(openWith);
	delete openWith;
    }
    else {
	// The file is no longer available.
	_fileUnavailMenu->Go(p, false, true, openRect);
    }
}


void
PieView::_launch(FileInfo *f, entry_ref *appRef)
{
    BMessage msg(B_REFS_RECEIVED);
    msg.AddRef("refs", &f->ref);

    if (appRef == NULL) {
	// Let the registrar pick an app based on the file's MIME type.
	BMimeType *type = f->type();
	be_roster->Launch(type->Type(), &msg);
	delete type;
    }
    else {
	// Launch a designated app to handle this file.
	be_roster->Launch(appRef, &msg);
    }
}


void
PieView::MessageReceived(BMessage *m)
{
    switch (m->what) {
	case B_SIMPLE_DATA:
	case B_REFS_RECEIVED:
	{
	    for (int i = 0; true; i++) {
		entry_ref ref;
		if (m->FindRef("refs", i, &ref) == B_OK) {
		    BEntry entry(&ref, true);
		    handleArg(&entry);
		}
		else {
		    break;
		}
	    }
	    break;
	}

	case kBtnRescan:
	    if (_currentVolume != NULL) {
		_scanners[_currentVolume]->refresh();
		Invalidate();
	    }
	    break;

	case kMenuSelectVol:
	{
	    BVolume *v;
	    m->FindPointer(kNameVolPtr, (void **)&v);
	    _showVolume(v);
	    break;
	}

	case kScanProgress:
	{
	    BVolume *v;
	    m->FindPointer(kNameVolPtr, (void **)&v);
	    if (v == _currentVolume) {
		// This is kind of a hack.  Instead of invalidating all or
		// part of the view, which tends to cause flicker as the view
		// is erased and redrawn, just call Draw() directly to update
		// the progress bar.
		Draw(BRect(0.0, 0.0, 0.0, 0.0));
	    }
	    break;
	}

	case kScanDone:
	{
	    BVolume *v;
	    m->FindPointer(kNameVolPtr, (void **)&v);
	    if (v == _currentVolume) {
		Invalidate();
	    }
	    break;
	}

        default:
            BView::MessageReceived(m);
            break;
    }
}


void
PieView::MouseDown(BPoint p)
{
    uint32 buttons;
    GetMouse(&p, &buttons);

    FileInfo *f = _fileAt(p);
    if (f != NULL && !f->pseudo) {
	if (buttons & B_PRIMARY_MOUSE_BUTTON) {
	    _clicked = true;
	    _whereClicked = p;
	    _fileClicked = f;
	    SetMouseEventMask(B_POINTER_EVENTS);
	}
	else if (buttons & B_SECONDARY_MOUSE_BUTTON) {
	    p = ConvertToScreen(p);
	    _menu(f, p);
	}
    }
}


void
PieView::MouseUp(BPoint p)
{
    // If the primary button was released and there was no dragging happening,
    // just zoom in or out.
    if (_clicked && !_dragging) {
	Scanner *scanner = _scanners[_currentVolume];
	FileInfo *f = _fileAt(p);
	if (f != NULL) {
	    if (f == scanner->currentDir()) {
		scanner->changeDir(f->parent);
		Invalidate();
	    }
	    else if (f->children.size() > 0) {
		scanner->changeDir(f);
		Invalidate();
	    }
	}
    }

    _clicked = false;
    _dragging = false;
}


void
PieView::MouseMoved(BPoint p, uint32 transit, const BMessage *m)
{
    if (_clicked) {
	// Primary mouse button is down.
	if (!_dragging) {
	    // If the mouse has moved far enough, initiate dragging.
	    BPoint diff = p - _whereClicked;
	    if (diff.x > kDragThreshold || diff.x < -kDragThreshold ||
		diff.y > kDragThreshold || diff.y < -kDragThreshold) {
		_dragging = true;

		BBitmap *icon = new BBitmap(BRect(0.0, 0.0, 31.0, 31.0), B_CMAP8);
		if (BNodeInfo::GetTrackerIcon(&_fileClicked->ref, icon, B_LARGE_ICON) == B_OK) {
		    BMessage msg(B_SIMPLE_DATA);
		    msg.AddRef("refs", &_fileClicked->ref);
		    DragMessage(&msg, icon, B_OP_BLEND, BPoint(15.0, 15.0));
		}
		else {
		    delete icon;
		}
	    }
	}
    }
    else {
	// Mouse button is not down, display file info.
	if (transit == B_EXITED_VIEW) {
	    // Clear status view
	    _win->statusView->show(NULL);
	}
	else {
	    // Display file information.
	    _win->statusView->show(_fileAt(p));
	}
    }
}


void
PieView::handleArg(BEntry *entry)
{
    // Going through these seemingly pointless gyrations (instead of getting
    // the dev_t directly from the BEntry) allows us to process volumes
    // properly.  It looks like a dropped volume refers to dev_t 1 (why?), but
    // getting the dev_t from the path allows us to get at the dev_t where the
    // root directory actually lives.
    BPath path;
    entry->GetPath(&path);
    dev_t dev = dev_for_path(path.Path());

    // Set the desired path in the scanner.
    Scanner *scanner = NULL;
    ScannerMap::iterator i = _scanners.begin();
    while (i != _scanners.end()) {
	if ((*i).second->device() == dev) {
	    scanner = (*i).second;
	    break;
	}
	i++;
    }
    if (scanner == NULL) {
	BVolume *v = _win->ctrlView->volCtrl->findDevice(dev);
	if (v != NULL) {
	    scanner = _scanners[v] = new Scanner(v, this);
	}
    }
    if (scanner != NULL) {
	string s(path.Path());
	scanner->setDesiredPath(s);

	// Select the volume on the menu.
	_win->ctrlView->volCtrl->findDevice(dev, true);
    }
}


void
PieView::Draw(BRect updateRect)
{
    if (_scanners.find(_currentVolume) != _scanners.end()) {
	// There is a current volume.
	Scanner *scanner = _scanners[_currentVolume];
	if (scanner->busy()) {
	    // Show progress of scanning.
	    _drawProgressBar(updateRect);
	    _win->ctrlView->enableRescan(false);
	}
	else {
	    _drawPieChart(updateRect);
	    _win->ctrlView->enableRescan(true);
	}
    }
    else {
	// No volume has been selected, so display a prompt for one.
	BRect b = Bounds();
	float strWidth = StringWidth(kVolPrompt);
	float bx = (b.left + b.Width() - strWidth) / 2.0;
	float by = (b.top + b.Height()) / 2.0;
	SetHighColor(0x00, 0x00, 0x00);
	DrawString(kVolPrompt, BPoint(bx, by));
	_win->ctrlView->enableRescan(false);
    }
}

