#include "scanner.h"
#include "common.h"


Scanner::Scanner(BVolume *v, BHandler *handler):
    BLooper(), _volume(v), _snapshot(NULL), _desiredPath(), _task(),
    _busy(false), _quitRequested(false)
{
    _messenger = new BMessenger(handler);

    _doneMsg = new BMessage(kScanDone);
    _doneMsg->AddPointer(kNameVolPtr, _volume);

    _progressMsg = new BMessage(kScanProgress);
    _progressMsg->AddPointer(kNameVolPtr, _volume);

    Run();
}


Scanner::~Scanner()
{
    delete _messenger;
    delete _doneMsg;
    delete _progressMsg;
    if (_snapshot != NULL) {
	delete _snapshot;
    }
}


void
Scanner::_refresh(FileInfo *startInfo)
{
    if (startInfo == NULL || startInfo == _snapshot->rootDir) {
	if (_snapshot != NULL) {
	    delete _snapshot;
	}
	_snapshot = new VolSnapshot(_volume);
	_task = kStrScanningX + _snapshot->name;
	_volBytesInUse = _snapshot->capacity - _snapshot->freeBytes;
	_volBytesScanned = 0;
	_progress = 0.0;
	_lastReport = -100.0;

	BDirectory root;
	_volume->GetRootDirectory(&root);
	_snapshot->rootDir = _getFileInfo(&root, NULL);

	FileInfo *fs = new FileInfo;
	fs->pseudo = true;
	string s = "Free on " + _snapshot->name;
	fs->ref.set_name(s.c_str());
	fs->size = _snapshot->freeBytes;
	_snapshot->freeSpace = fs;

	_snapshot->currentDir = NULL;

    }
    else {
	_snapshot->capacity = _volume->Capacity();
	_snapshot->freeBytes = _volume->FreeBytes();
	_task = kStrScanningX + string(startInfo->ref.name);
	_volBytesInUse = _snapshot->capacity - _snapshot->freeBytes;
	_volBytesScanned = _volBytesInUse - startInfo->size; // best guess
	_progress = 100.0 * _volBytesScanned / _volBytesInUse;
	_lastReport = -100.0;

	BDirectory startDir(&startInfo->ref);
	if (startDir.InitCheck() == B_OK) {
	    FileInfo *parent = startInfo->parent;
	    vector<FileInfo *>::iterator i = parent->children.begin();
	    FileInfo *newInfo = _getFileInfo(&startDir, parent);
	    while (i != parent->children.end() && *i != startInfo) {
		i++;
	    }
	    int idx = i - parent->children.begin();
	    parent->children[idx] = newInfo;

	    // Fixup count and size fields in parent directory.
	    parent->size += newInfo->size - startInfo->size;
	    parent->count += newInfo->count - startInfo->count;

	    delete startInfo;
	}
    }


    _busy = false;
    _changeToDesired();
    _messenger->SendMessage(_doneMsg);
}


FileInfo *
Scanner::_getFileInfo(BDirectory *dir, FileInfo *parent)
{
    FileInfo *thisDir = new FileInfo;
    BEntry entry;
    dir->GetEntry(&entry);
    entry.GetRef(&thisDir->ref);
    thisDir->parent = parent;
    thisDir->count = 0;

    while (true) {
	if (_quitRequested) {
	    break;
	}
	if (dir->GetNextEntry(&entry) == B_ENTRY_NOT_FOUND) {
	    break;
	}
	if (entry.IsSymLink()) {
	    continue;
	}

	if (entry.IsFile()) {
	    FileInfo *child = new FileInfo;
	    entry.GetRef(&child->ref);
	    entry.GetSize(&child->size);
	    child->parent = thisDir;
	    thisDir->children.push_back(child);

	    // Send a progress report periodically.
	    _volBytesScanned += child->size;
	    _progress = 100.0 * _volBytesScanned / _volBytesInUse;
	    if (_progress - _lastReport > kReportInterval) {
		_lastReport = _progress;
		_messenger->SendMessage(_progressMsg);
	    }
	}
	else if (entry.IsDirectory()) {
	    BDirectory childDir(&entry);
	    thisDir->children.push_back(_getFileInfo(&childDir, thisDir));
	}
	thisDir->count++;
    }

    vector<FileInfo *>::iterator i = thisDir->children.begin();
    while (i != thisDir->children.end()) {
	thisDir->size += (*i)->size;
	thisDir->count += (*i)->count;
	i++;
    }

    return thisDir;
}


void
Scanner::_changeToDesired()
{
    if (_busy) {
	return;
    }

    if (_desiredPath.size() > 0) {
	char *workPath = strdup(_desiredPath.c_str());
	char *pathPtr = &workPath[1]; // skip leading '/'
	char *endOfPath = pathPtr + strlen(pathPtr);

	// Check the name of the root directory.  It is a special case because
	// it has no parent data structure.
	FileInfo *checkDir = _snapshot->rootDir;
	FileInfo *prefDir = NULL;
	char *sep = strchr(pathPtr, '/');
	if (sep != NULL) {
	    *sep = '\0';
	}
	if (strcmp(pathPtr, checkDir->ref.name) == 0) {
	    // Root directory matches, so follow the remainder of the path as
	    // far as possible.
	    prefDir = checkDir;
	    pathPtr += 1 + strlen(pathPtr);
	    while (pathPtr < endOfPath) {
		sep = strchr(pathPtr, '/');
		if (sep != NULL) {
		    *sep = '\0';
		}
		checkDir = prefDir->findChild(pathPtr);
		if (checkDir == NULL || checkDir->children.size() == 0) {
		    break;
		}
		pathPtr += 1 + strlen(pathPtr);
		prefDir = checkDir;
	    }
	}

	// If the desired path is the volume's root directory, default to the
	// volume display.
	if (prefDir == _snapshot->rootDir) {
	    prefDir = NULL;
	}
	changeDir(prefDir);

	free(workPath);
	_desiredPath.erase();
    }
}


void
Scanner::refresh(FileInfo *startInfo)
{
    if (_busy) {
	return;
    }

    _busy = true;

    // Remember the current directory, if any, so we can return to it after
    // the scanning is done.
    if (_snapshot != NULL && _snapshot->currentDir != NULL) {
	_snapshot->currentDir->getPath(_desiredPath);
    }

    _task.assign(kEmptyStr);

    BMessage msg(kScanRefresh);
    msg.AddPointer(kNameFilePtr, startInfo);
    PostMessage(&msg);
}


void
Scanner::setDesiredPath(string &path)
{
    _desiredPath = path;

    if (_snapshot == NULL) {
	refresh();
    }
    else {
	_changeToDesired();
    }
}


void
Scanner::requestQuit()
{
    _quitRequested = true;
    Lock();
    Quit();
}


void
Scanner::MessageReceived(BMessage *m)
{
    switch (m->what) {
	case kScanRefresh:
	{
	    FileInfo *startInfo;
	    m->FindPointer(kNameFilePtr, (void **)&startInfo);
	    _refresh(startInfo);
	    break;
	}

	default:
	    BLooper::MessageReceived(m);
	    break;
    }
}

